from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, Hashable, List, Optional, Sequence, Set, Tuple, Type, TypeVar

import colorful as cf
import networkx as nx
from flax.serialization import from_state_dict, register_serialization_state, to_state_dict
from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import BatchDimensionIterator, display_class, display_generic, display_key_and_value

__all__: List[str] = []

T = TypeVar('T', bound=nx.Graph)


def register_graph_as_jax_pytree(cls: Type[T]) -> None:
    def tree_unflatten(hashed: Hashable, trees: Sequence[PyTree]) -> T:
        node_dicts, edge_dicts = trees

        if not isinstance(node_dicts, dict):
            raise TypeError
        if not isinstance(edge_dicts, dict):
            raise TypeError

        graph = cls()
        graph.add_nodes_from(node_dicts.items())
        graph.add_edges_from([(source, target, data)
                              for (source, target), data in edge_dicts.items()])

        return graph

    def tree_flatten(graph: T) -> Tuple[Sequence[PyTree], Hashable]:
        return ((dict(graph.nodes), dict(graph.edges)), None)

    register_pytree_node(cls, tree_flatten, tree_unflatten)


register_graph_as_jax_pytree(nx.Graph)
register_graph_as_jax_pytree(nx.DiGraph)


def register_graph_as_flax_state_dict(cls: Type[T]) -> None:
    def ty_to_state_dict(graph: T) -> Dict[str, Any]:
        edge_dict_of_dicts = defaultdict[Any, Dict[Any, Any]](dict)
        for (source, target), edge_dict in dict(graph.edges).items():
            edge_dict_of_dicts[source][target] = edge_dict
        return {'nodes': to_state_dict(dict(graph.nodes)),
                'edges': to_state_dict(dict(edge_dict_of_dicts))}

    def ty_from_state_dict(graph: T, state_dict: Dict[str, Any]) -> T:
        retval = type(graph)()
        for node_name, node_dict in state_dict['nodes'].items():
            retval.add_node(node_name, **from_state_dict(graph.nodes[node_name], node_dict))
        for source, target_and_edge_dict in state_dict['edges'].items():
            for target, edge_dict in target_and_edge_dict.items():
                retval.add_edge(source, target, **from_state_dict(graph.edges[source, target],
                                                                  edge_dict))
        return retval

    register_serialization_state(cls, ty_to_state_dict,  # type: ignore[no-untyped-call]
                                 ty_from_state_dict)


register_graph_as_flax_state_dict(nx.Graph)
register_graph_as_flax_state_dict(nx.DiGraph)


@display_generic.register
def _(value: nx.Graph,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    directed = isinstance(value, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(value))
    bdi = BatchDimensionIterator(batch_dims)
    for name, node in value.nodes.items():
        sub_batch_dims = bdi.advance(node)
        retval += display_key_and_value(name, node, ": ", seen, show_values, indent + 1,
                                        sub_batch_dims)
    for (source, target), edge in value.edges.items():
        key = f"{source}{arrow}{target}"
        sub_batch_dims = bdi.advance(edge)
        retval += display_key_and_value(key, edge, ": ", seen, show_values, indent + 1,
                                        sub_batch_dims)
    return retval



from ._src.dataclasses.dataclass import dataclass
from ._src.dataclasses.dataclass_transform import dataclass_transform
from ._src.dataclasses.helpers import as_shallow_dict, field

__all__ = ['dataclass', 'dataclass_transform', 'as_shallow_dict', 'field']



from __future__ import annotations

from reprlib import recursive_repr
from typing import (Any, Callable, Dict, Generic, Hashable, Mapping, Optional, Sequence, Tuple,
                    TypeVar, Union, cast)

from jax.tree_util import register_pytree_node_class

from .annotations import PyTree

__all__ = ['Partial']


R = TypeVar('R')


@register_pytree_node_class
class Partial(Generic[R]):
    """
    A version of functools.partial that returns a pytree.

    Use it for partial function evaluation in a way that is compatible with JAX's transformations,
    e.g., ``Partial(func, *args, **kwargs)``.

    JAX's version of this class in jax.tree_util assumes that all parameters are dynamic and the
    callable is static.  This version allows you to specify which parameters are static and whether
    the callable is static.
    """
    def __init__(self,
                 function: Union[Partial[R], Callable[..., R]],
                 /,
                 *args: Any,
                 callable_is_static: bool = True,
                 static_argnums: Tuple[int, ...] = (),
                 static_kwargs: Optional[Mapping[str, Any]] = None,
                 **dynamic_kwargs: Any) -> None:
        """
        Args:
            function: The function being applied.
            args: The static and dynamic applied positional arguments.
            callable_is_static: Whether the function callable is static.
            static_argnums: The indices of the applied positional arguments that are static.
            static_kwargs: The static applied keyword arguments.
            dynamic_kwargs: The dynamic applied keyword arguments.
        """
        super().__init__()
        if isinstance(function, Partial):
            if callable_is_static:
                raise TypeError
            # Could collapse the arguments here, but the benefit is small.
        self.function = function
        self.callable_is_static = callable_is_static
        self.args = args
        self.static_argnums = static_argnums
        self.static_kwargs = {} if static_kwargs is None else static_kwargs
        self.dynamic_kwargs = dynamic_kwargs

    def tree_flatten(self: Partial[R]) -> Tuple[Sequence[PyTree], Hashable]:
        static_args, dynamic_args = self._partition_args()
        static_kwargs = tuple((key, self.static_kwargs[key]) for key in sorted(self.static_kwargs))
        return ((dynamic_args, self.dynamic_kwargs),
                (self.callable_is_static, self.static_argnums, static_args, static_kwargs))

    @classmethod
    def tree_unflatten(cls,
                       static: Hashable,
                       trees: Sequence[PyTree]) -> Partial[R]:
        if not isinstance(static, tuple):
            raise RuntimeError

        callable_is_static, static_argnums, static_args, static_kwarg_items = static
        static_kwargs = dict(static_kwarg_items)

        if not isinstance(static_args, tuple):
            raise RuntimeError

        dynamic_args, dynamic_kwargs = trees

        if not isinstance(dynamic_args, tuple):
            raise RuntimeError
        if not isinstance(dynamic_kwargs, dict):
            raise RuntimeError

        dynamic_kwargs = cast(Dict[str, Any], dynamic_kwargs)
        args = cls._unpartition_args(callable_is_static, static_argnums, static_args, dynamic_args)

        return Partial[R](*args,
                          callable_is_static=callable_is_static,
                          static_argnums=static_argnums,
                          static_kwargs=static_kwargs,
                          **dynamic_kwargs)

    # Magic methods --------------------------------------------------------------------------------
    def __call__(self, *args: Any, **kwargs: Any) -> R:
        keywords = {**self.static_kwargs, **self.dynamic_kwargs, **kwargs}
        return self.function(*self.args, *args, **keywords)  # pyright: ignore

    @recursive_repr()
    def __repr__(self) -> str:
        qualname = type(self).__qualname__
        args = [repr(self.function)]
        args.extend(repr(x) for x in self.args)
        args.append(f"callable_is_static={self.callable_is_static}")
        args.append(f"static_argnums={self.static_argnums}")
        args.append(f"static_kwargs={self.static_kwargs}")
        args.extend(f"{k}={v!r}" for (k, v) in self.dynamic_kwargs.items())
        return f"{qualname}({', '.join(args)})"

    # Private methods ------------------------------------------------------------------------------
    def _partition_args(self) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
        # Partition self.args into static and dynamic arguments.
        static_args = []
        dynamic_args = []

        static_argnums = set(self.static_argnums)

        def _append(is_static: bool, value: Any) -> None:
            if is_static:
                static_args.append(value)
            else:
                dynamic_args.append(value)

        _append(self.callable_is_static, self.function)
        for i, value in enumerate(self.args):
            _append(i in static_argnums, value)

        return tuple(reversed(static_args)), tuple(reversed(dynamic_args))

    @classmethod
    def _unpartition_args(cls,
                          callable_is_static: bool,
                          static_argnums: Tuple[int, ...],
                          static_args: Tuple[Any, ...],
                          dynamic_args: Tuple[Any, ...]) -> Tuple[Any, ...]:
        static_arg_list = list(static_args)
        dynamic_arg_list = list(dynamic_args)
        args = []
        for i in range(len(static_args) + len(dynamic_args)):
            is_static = (callable_is_static
                         if i == 0
                         else i - 1 in static_argnums)
            if is_static:
                args.append(static_arg_list.pop())
            else:
                args.append(dynamic_arg_list.pop())
        return tuple(args)



from __future__ import annotations

from typing import Any, Optional, Type

__all__ = ['dataclass_transform']


def dataclass_transform(cls: Optional[Type[Any]] = None, /, *, init: bool = True, repr_: bool =
                        True, eq: bool = True, order: bool = False) -> Any:
    return cls



from __future__ import annotations

from typing import Any, List, Optional, Type, TypeVar, Union

import jax.numpy as jnp
import jax.random
import numpy as np
from jax.random import KeyArray, PRNGKey  # pyright: ignore

from .annotations import Array, BooleanArray, RealArray, RealNumeric, Shape, ShapeLike
from .dataclasses import dataclass

__all__ = ['Generator']


T = TypeVar('T', bound='Generator')
DTypeLikeFloat = Any


@dataclass
class Generator:
    """
    This class represents a JAX random number generator.  Unlike `numpy.Generator`, `tjax.Generator`
    has no mutating methods.  Instead, its generation methods return a new instance along with the
    generated tensor.
    """
    key: KeyArray

    # Class methods --------------------------------------------------------------------------------
    @classmethod
    def from_seed(cls: Type[T], seed: int) -> T:
        return cls(PRNGKey(seed))

    # New methods ----------------------------------------------------------------------------------
    def is_vmapped(self) -> bool:
        return self.key.shape != (2,)

    def split(self, n: int = 2) -> List[Generator]:
        """
        Split a generator into a list of generators.
        """
        assert not self.is_vmapped()
        keys = jax.random.split(self.key, n)
        return [Generator(key=key) for key in keys]

    def vmap_split(self, shape: ShapeLike) -> Generator:
        """
        Split a generator into a generator that can be passed to a vmapped function.
        """
        assert not self.is_vmapped()
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        prod_shape = int(np.prod(shape))
        keys = (self.key
                if prod_shape == 1
                else jax.random.split(self.key, prod_shape))
        # https://github.com/google/jax/issues/6473
        keys = keys.reshape(shape + (2,))
        return Generator(keys)

    def bernoulli(self, p: RealNumeric, shape: Optional[Shape] = None) -> BooleanArray:
        return jax.random.bernoulli(self.key, p, shape)  # type: ignore[return-value]

    def choice(self,
               a: Union[int, Array],
               shape: Shape = (),
               replace: bool = True,
               p: Optional[RealArray] = None) -> Union[int, Array]:
        return jax.random.choice(self.key, a, shape, replace, p)  # type: ignore[return-value]

    def gamma(self,
              gamma_shape: RealNumeric,
              shape: Optional[Shape] = None,
              dtype: DTypeLikeFloat = np.float64) -> RealArray:
        return jax.random.gamma(self.key, gamma_shape, shape)  # type: ignore[return-value]

    def normal(self,
               shape: Shape = (),
               dtype: DTypeLikeFloat = np.float64) -> jnp.ndarray:
        return jax.random.normal(self.key, shape, dtype)

    def uniform(self,
                shape: Shape = (),
                dtype: DTypeLikeFloat = np.float64,
                minval: RealNumeric = 0.,
                maxval: RealNumeric = 1.) -> RealArray:
        return jax.random.uniform(self.key,  # type: ignore[return-value]
                                  shape, dtype, minval, maxval)
